#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdint.h>
#include <stdlib.h>	

#define SPECIFY_DEVICE				0
#define I_O_REGISTER_DEFINITIONS	1
#define SEC_BIT_DEFS				2
#define SEC_LOCKBITS				3
#define DATA_MEMORY_DECLARATIONS	4
#define BOOTLOADER_DECLARATIONS		5
#define INTERRUPT_VECTORS			6

typedef struct port{
	port *next;
	char name[20];
	int adr;
	int unused;
} port;

int section=-1;
FILE *dst;

port *root, *last;

int srchPnum(char *name){
	port *act;
	act=root;
	while(act){
		if(!strcmp(act->name,name)){
			act->unused=0;
			return act->adr;
		}
		act=act->next;
	}
	return -1;
}

void procLine(char *st){
	int i,c;
	char *cp,*cp1,*cp2,st1[2000];
	port *act,*next;
	if(cp=strchr(st,'\r')) *cp=0;
	if(cp=strchr(st,'\n')) *cp=0;
	cp=st;
	while(*cp){
		if(*cp=='/') *cp='_';
		cp++;
	}

	if((!strncmp(st,"; ***** ",8))&&(strlen(st)>39)){
		/* Sektion */
		cp=st+strlen(st)-3;
		while((*cp)=='*') cp--;
		*cp=0;
		//printf("Sektion <%s>\n",st+8);
		if(!strcmp(st+8,"SPECIFY DEVICE")) section=SPECIFY_DEVICE;
		if(!strcmp(st+8,"I_O REGISTER DEFINITIONS")) section=I_O_REGISTER_DEFINITIONS;
		if(!strcmp(st+8,"BIT DEFINITIONS")){
			section=SEC_BIT_DEFS;
			fprintf(dst,"[PortDefs]\n");
		}

		if(!strcmp(st+8,"LOCKSBITS")){
			section=SEC_LOCKBITS;
			fprintf(dst,"\nGroup Orphaned_Ports\n");
			act=root;
			while(act){
				if(act->unused) fprintf(dst," Port %s=%02X\n",act->name,act->adr);
				next=act->next;
				free(act);
				act=next;
			}
			root=0;
		}

		if(!strcmp(st+8,"DATA MEMORY DECLARATIONS")){
			section=DATA_MEMORY_DECLARATIONS;
			fprintf(dst,"\n[MemoryLimits]\n");
		}

		if(!strcmp(st+8,"BOOTLOADER DECLARATIONS")){
			section=BOOTLOADER_DECLARATIONS;
		}

		if(!strcmp(st+8,"INTERRUPT VECTORS")){
			section=INTERRUPT_VECTORS;
			fprintf(dst,"\n[InterruptVectors]\n");
		}

		return;
	}

	if((!strncmp(st,"; ***** ",8))&&(strlen(st)==38)){
		cp=st+8;
		while(!isblank(*cp)) cp++;
		*cp=0;
		fprintf(dst,"\nGroup %s\n",st+8);
		return;
	}
	switch(section){

		case SPECIFY_DEVICE:
		if(!strncmp(st,".device ",8)){
			fprintf(dst,"Device=%s\n\n",st+8);
		}
		break;

		case I_O_REGISTER_DEFINITIONS:
		if(strncmp(st,".equ",4)) return;
		cp=st+5;
		cp1=cp;
		while((isalnum(*cp1))||(*cp1=='_')) cp1++;
		*cp1=0;
		cp1++;
		cp1=strstr(cp1,"0x");
		if(!cp1) return;
		cp1+=2;
		sscanf(cp1,"%x",&i);
		if(!strstr(cp1,"MEMORY MAPPED")) i+=32;
		act=(port*)calloc(sizeof(port),1);
		strcpy(act->name,cp);
		act->adr=i;
		act->unused=1;
		if(root) last->next=act;
		else root=act;
		last=act;
		//printf("%s:%02x ;%s\n",cp,i,cp1);
		break;

		case SEC_BIT_DEFS:
		if(!strncmp(st,"; ",2)){
			cp1=strstr(st," - ");
			if(!cp1) return;
			*cp1=0;
			cp1+=3;
			i=srchPnum(st+2);
			fprintf(dst," Port %s=%02X ;%s\n",st+2,i,cp1);
		}
		if(!strncmp(st,".equ",4)){
			if(strstr(st,"compatibility")) return;
			cp1=st+5;
			while(isalnum(*cp1)) cp1++;
			*cp1=0;
			cp1+=3;
			if(!isdigit(*cp1)) return;
			i=atoi(cp1);
			cp1+=2;
			fprintf(dst,"  Bit %s=%d %s\n",st+5,i,cp1);
		}
		return;
		break;

		case DATA_MEMORY_DECLARATIONS:
		case INTERRUPT_VECTORS:
		if(strncmp(st,".equ",4)) return;
		cp=st+5;
		while((isalnum(*cp))||(*cp=='_')) cp++;
		*cp=0;
		cp+=3;
		fprintf(dst,"%s=%s\n",st+5,cp);
		break;
	}
}



void procFile(char *fn){
	FILE *f;
	char st[2048],*cp;
	section=-1;
	printf("%s\n",fn);
	strcpy(st,fn);
	cp=strstr(st,"def.inc");
	if(!cp) return;
	strcpy(cp,".def");
	f=fopen(fn,"r");
	if(!f) return;
	dst=fopen(st,"w");
	while(fgets(st,2040,f)) procLine(st);
	fclose(f);
	fclose(dst);
}


int main(int argc, char **argv){
	int c;
	for(c=1;c<argc;c++) procFile(argv[c]);
	return 0;
}